package com.sun.corba.se.impl.presentation.rmi;

import org.omg.CORBA.portable.ApplicationException;
import com.sun.corba.se.spi.orb.ORB;
import javax.rmi.CORBA.Util;
import javax.rmi.PortableRemoteObject;
import org.omg.CORBA_2_3.portable.OutputStream;
import org.omg.CORBA_2_3.portable.InputStream;
import java.rmi.Remote;
import java.rmi.RemoteException;
import org.omg.CORBA.portable.IDLEntity;
import java.io.Externalizable;
import java.io.Serializable;
import java.lang.reflect.Method;
import com.sun.corba.se.spi.presentation.rmi.DynamicMethodMarshaller;

public class DynamicMethodMarshallerImpl implements DynamicMethodMarshaller
{
    Method method;
    ExceptionHandler ehandler;
    boolean hasArguments;
    boolean hasVoidResult;
    boolean needsArgumentCopy;
    boolean needsResultCopy;
    DynamicMethodMarshallerImpl.ReaderWriter[] argRWs;
    DynamicMethodMarshallerImpl.ReaderWriter resultRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter booleanRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter byteRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter charRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter shortRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter intRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter longRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter floatRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter doubleRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter corbaObjectRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter anyRW;
    private static DynamicMethodMarshallerImpl.ReaderWriter abstractInterfaceRW;
    
    private static boolean isAnyClass(Class clazz) {
        return clazz.equals(Object.class) || clazz.equals(Serializable.class) || clazz.equals(Externalizable.class);
    }
    
    private static boolean isAbstractInterface(Class clazz) {
        if (IDLEntity.class.isAssignableFrom(clazz)) {
            return clazz.isInterface();
        }
        return clazz.isInterface() && allMethodsThrowRemoteException(clazz);
    }
    
    private static boolean allMethodsThrowRemoteException(Class clazz) {
        Method[] methods = clazz.getMethods();
        int i = 0;
        while (i < methods.length) {
            Method method = methods[i];
            if (method.getDeclaringClass() != Object.class && !throwsRemote(method)) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    private static boolean throwsRemote(Method method) {
        Class[] exceptionTypes = method.getExceptionTypes();
        int i = 0;
        while (i < exceptionTypes.length) {
            if (RemoteException.class.isAssignableFrom(exceptionTypes[i])) {
                return true;
            }
            ++i;
        }
        return false;
    }
    
    public static DynamicMethodMarshallerImpl.ReaderWriter makeReaderWriter(final Class clazz) {
        if (clazz.equals(Boolean.TYPE)) {
            return DynamicMethodMarshallerImpl.booleanRW;
        }
        if (clazz.equals(Byte.TYPE)) {
            return DynamicMethodMarshallerImpl.byteRW;
        }
        if (clazz.equals(Character.TYPE)) {
            return DynamicMethodMarshallerImpl.charRW;
        }
        if (clazz.equals(Short.TYPE)) {
            return DynamicMethodMarshallerImpl.shortRW;
        }
        if (clazz.equals(Integer.TYPE)) {
            return DynamicMethodMarshallerImpl.intRW;
        }
        if (clazz.equals(Long.TYPE)) {
            return DynamicMethodMarshallerImpl.longRW;
        }
        if (clazz.equals(Float.TYPE)) {
            return DynamicMethodMarshallerImpl.floatRW;
        }
        if (clazz.equals(Double.TYPE)) {
            return DynamicMethodMarshallerImpl.doubleRW;
        }
        if (Remote.class.isAssignableFrom(clazz)) {
            return new DynamicMethodMarshallerImpl.ReaderWriterBase("remote(" + clazz.getName() + ")") {
                @Override
                public Object read(InputStream inputStream) {
                    return PortableRemoteObject.narrow(inputStream.read_Object(), clazz);
                }
                
                @Override
                public void write(OutputStream outputStream, Object o) {
                    Util.writeRemoteObject(outputStream, o);
                }
            };
        }
        if (clazz.equals(org.omg.CORBA.Object.class)) {
            return DynamicMethodMarshallerImpl.corbaObjectRW;
        }
        if (org.omg.CORBA.Object.class.isAssignableFrom(clazz)) {
            return new DynamicMethodMarshallerImpl.ReaderWriterBase("org.omg.CORBA.Object(" + clazz.getName() + ")") {
                @Override
                public Object read(InputStream inputStream) {
                    return inputStream.read_Object(clazz);
                }
                
                @Override
                public void write(OutputStream outputStream, Object o) {
                    outputStream.write_Object((org.omg.CORBA.Object)o);
                }
            };
        }
        if (isAnyClass(clazz)) {
            return DynamicMethodMarshallerImpl.anyRW;
        }
        if (isAbstractInterface(clazz)) {
            return DynamicMethodMarshallerImpl.abstractInterfaceRW;
        }
        return new DynamicMethodMarshallerImpl.ReaderWriterBase("value(" + clazz.getName() + ")") {
            @Override
            public Object read(InputStream inputStream) {
                return inputStream.read_value(clazz);
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                outputStream.write_value((Serializable)o, clazz);
            }
        };
    }
    
    public DynamicMethodMarshallerImpl(Method method) {
        this.hasArguments = true;
        this.hasVoidResult = true;
        this.argRWs = null;
        this.resultRW = null;
        this.method = method;
        this.ehandler = new ExceptionHandlerImpl(method.getExceptionTypes());
        this.needsArgumentCopy = false;
        Class[] parameterTypes = method.getParameterTypes();
        this.hasArguments = (parameterTypes.length > 0);
        if (this.hasArguments) {
            this.argRWs = new DynamicMethodMarshallerImpl.ReaderWriter[parameterTypes.length];
            for (int i = 0; i < parameterTypes.length; this.argRWs[i] = makeReaderWriter(parameterTypes[i]), ++i) {
                if (!parameterTypes[i].isPrimitive()) {
                    this.needsArgumentCopy = true;
                }
            }
        }
        Class<Object> returnType = method.getReturnType();
        this.needsResultCopy = false;
        if (!(this.hasVoidResult = returnType.equals(Void.TYPE))) {
            this.needsResultCopy = !returnType.isPrimitive();
            this.resultRW = makeReaderWriter(returnType);
        }
    }
    
    public Method getMethod() {
        return this.method;
    }
    
    public Object[] copyArguments(Object[] array, ORB orb) throws RemoteException {
        if (this.needsArgumentCopy) {
            return Util.copyObjects(array, orb);
        }
        return array;
    }
    
    public Object[] readArguments(InputStream inputStream) {
        Object[] array = null;
        if (this.hasArguments) {
            array = new Object[this.argRWs.length];
            for (int i = 0; i < this.argRWs.length; ++i) {
                array[i] = this.argRWs[i].read(inputStream);
            }
            return array;
        }
        return array;
    }
    
    public void writeArguments(OutputStream outputStream, Object[] array) {
        if (this.hasArguments) {
            if (array.length != this.argRWs.length) {
                throw new IllegalArgumentException("Expected " + this.argRWs.length + " arguments, but got " + array.length + " arguments.");
            }
            for (int i = 0; i < this.argRWs.length; ++i) {
                this.argRWs[i].write(outputStream, array[i]);
            }
        }
    }
    
    public Object copyResult(Object o, ORB orb) throws RemoteException {
        if (this.needsResultCopy) {
            return Util.copyObject(o, orb);
        }
        return o;
    }
    
    public Object readResult(InputStream inputStream) {
        if (this.hasVoidResult) {
            return null;
        }
        return this.resultRW.read(inputStream);
    }
    
    public void writeResult(OutputStream outputStream, Object o) {
        if (!this.hasVoidResult) {
            this.resultRW.write(outputStream, o);
        }
    }
    
    public boolean isDeclaredException(Throwable t) {
        return this.ehandler.isDeclaredException(t.getClass());
    }
    
    public void writeException(OutputStream outputStream, Exception ex) {
        this.ehandler.writeException(outputStream, ex);
    }
    
    public Exception readException(ApplicationException ex) {
        return this.ehandler.readException(ex);
    }
    
    static {
        DynamicMethodMarshallerImpl.booleanRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("boolean") {
            @Override
            public Object read(InputStream inputStream) {
                return new Boolean(inputStream.read_boolean());
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                outputStream.write_boolean((boolean)o);
            }
        };
        DynamicMethodMarshallerImpl.byteRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("byte") {
            @Override
            public Object read(InputStream inputStream) {
                return new Byte(inputStream.read_octet());
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                outputStream.write_octet((byte)o);
            }
        };
        DynamicMethodMarshallerImpl.charRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("char") {
            @Override
            public Object read(InputStream inputStream) {
                return new Character(inputStream.read_wchar());
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                outputStream.write_wchar((char)o);
            }
        };
        DynamicMethodMarshallerImpl.shortRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("short") {
            @Override
            public Object read(InputStream inputStream) {
                return new Short(inputStream.read_short());
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                outputStream.write_short((short)o);
            }
        };
        DynamicMethodMarshallerImpl.intRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("int") {
            @Override
            public Object read(InputStream inputStream) {
                return new Integer(inputStream.read_long());
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                outputStream.write_long((int)o);
            }
        };
        DynamicMethodMarshallerImpl.longRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("long") {
            @Override
            public Object read(InputStream inputStream) {
                return new Long(inputStream.read_longlong());
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                outputStream.write_longlong((long)o);
            }
        };
        DynamicMethodMarshallerImpl.floatRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("float") {
            @Override
            public Object read(InputStream inputStream) {
                return new Float(inputStream.read_float());
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                outputStream.write_float((float)o);
            }
        };
        DynamicMethodMarshallerImpl.doubleRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("double") {
            @Override
            public Object read(InputStream inputStream) {
                return new Double(inputStream.read_double());
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                outputStream.write_double((double)o);
            }
        };
        DynamicMethodMarshallerImpl.corbaObjectRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("org.omg.CORBA.Object") {
            @Override
            public Object read(InputStream inputStream) {
                return inputStream.read_Object();
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                outputStream.write_Object((org.omg.CORBA.Object)o);
            }
        };
        DynamicMethodMarshallerImpl.anyRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("any") {
            @Override
            public Object read(InputStream inputStream) {
                return Util.readAny(inputStream);
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                Util.writeAny(outputStream, o);
            }
        };
        DynamicMethodMarshallerImpl.abstractInterfaceRW = new DynamicMethodMarshallerImpl.ReaderWriterBase("abstract_interface") {
            @Override
            public Object read(InputStream inputStream) {
                return inputStream.read_abstract_interface();
            }
            
            @Override
            public void write(OutputStream outputStream, Object o) {
                Util.writeAbstractObject(outputStream, o);
            }
        };
    }
    
    abstract static class ReaderWriterBase implements DynamicMethodMarshallerImpl.ReaderWriter
    {
        String name;
        
        public ReaderWriterBase(String name) {
            this.name = name;
        }
        
        @Override
        public String toString() {
            return "ReaderWriter[" + this.name + "]";
        }
    }
    
    public interface ReaderWriter
    {
        public abstract Object read(InputStream p0);
        
        public abstract void write(OutputStream p0, Object p1);
    }
}
